/* Demonstration program showing use of the Array package and         */
/* Complex class.                                                     */
/*                                                                    */
/* CFFT: This program computes a two-dimensional FFT.                 */
/*                                                                    */
/* Results are checked by performing a forward FFT, then a reverse    */
/* FFT, and comparing the result with the original data.              */

import com.ibm.math.Complex;	// import the Complex class
import com.ibm.math.array.*;	// import the Array package

public class cfft {

  public static void main(String[] args) {

    int n = 256;	// problem size
    int reps = 1;	// number of repetitions of FFT pairs

    /*
     * Create problem arrays. FFT will be computed on array A.
     * Array B will store original value for comparison.
     */
    ComplexArray2D a = new ComplexArray2D(n,n);
    ComplexArray2D b = new ComplexArray2D(n,n);

    int i;
    int j;
    double t1;
    double t2;
    double time;
    double error;
    double ops;
    double mflops;
    double eps;

    eps = 1.0e-9;

    /*
     * Initialize the data.
     */
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        a.set(j,i,(new Complex(1.0,1.0)));
        b.set(j,i,a.get(j,i));
      };
    };
    
    /*
     * Peform the forward and reverse FFTs.
     */
    t1 = (1e-3*System.currentTimeMillis());
    for( i = 1; i <= reps; i++) {
      fft2d(a,n,1);
      fft2d(a,n,( -(1)));
    };
    t2 = (1e-3*System.currentTimeMillis());
    time = (t2 - t1);

    /*
     * Compute error between result and original data.
     */
    error = 0.0;
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        error = (error + (Complex.abs((a.get(j,i).minus(b.get(j,i))))));
      };
    };
    if (error > eps) {
      System.out.println("Result is incorrect!");
    } else {
      System.out.println("Result is correct!");
    }

    /*
     * Print performance results.
     */
    ops = (((((4.0 * reps) * n) * n) * ((int)((0.5+(1.0/Math.log(2))*Math.log(((double)(n))))))) * 10.0);
    mflops = ((1.0e-6 * ops) / time);
    System.out.print("Problem size: "+n+"x"+n);
    System.out.println(" Time: "+time+" Mflops: "+mflops);
  }

  public static void fft2d(ComplexArray2D a,int n,int isign) {
    if ((isign == 1)) {
      fft1drow(a,n,isign);
      fft1dcol(a,n,isign);
    };
    if ((isign == ( -(1)))) {
      fft1dcol(a,n,isign);
      fft1drow(a,n,isign);
    };
  }

  public static void fft1dcol(ComplexArray2D a,int n,int isign) {
    double pi = 3.14159265358979324;
    int log2n;
    int remainder;
    int point;
    int reverse;
    int row;
    int col;
    int i;
    int j;
    int k;
    int l;
    int stage;
    int lstar;
    Complex temp = new Complex(0.0,0.0);
    Complex tau = new Complex(0.0,0.0);
    Complex w = new Complex(0.0,0.0);
    Complex wp = new Complex(0.0,0.0);
    double theta;
    double s;
    temp.assign((new Complex(0.0,0.0)));
    tau.assign((new Complex(0.0,0.0)));
    w.assign((new Complex(0.0,0.0)));
    wp.assign((new Complex(0.0,0.0)));
    log2n = ((int)((0.5+(1.0/Math.log(2))*Math.log(((double)(n))))));
    for( col = 0; col <= (n - 1); col++) {
      for( point = 0; point <= (n - 1); point++) {
        remainder = point;
        reverse = 0;
        for( i = 1; i <= log2n; i++) {
          reverse = ((reverse * 2) + (remainder % 2));
          remainder = (remainder / 2);
        };
        if ((reverse > point)) {
          temp.assign(a.get(col,point));
          a.set(col,point,a.get(col,reverse));
          a.set(col,reverse,temp);
        };
      };
      l = 1;
      for( stage = 1; stage <= log2n; stage++) {
        lstar = l;
        l = (l * 2);
        theta = (((2 * pi) * isign) / l);
        wp.assign((new Complex(((( -(2.0)) * (Math.sin((0.5 * theta)))) * (Math.sin((0.5 * theta)))),(Math.sin(theta)))));
        w.assign((new Complex(1.0,0.0)));
        for( k = 0; k <= (lstar - 1); k++) {
          for( i = k; i <= (n - 1); i += l) {
            j = (i + lstar);
            tau.assign((a.get(col,j).times(w)));
            a.set(col,j,(a.get(col,i).minus(tau)));
            a.set(col,i,(a.get(col,i).plus(tau)));
          };
          w.assign((w.plus((w.times(wp)))));
        };
      };
    };
    if ((isign == ( -(1)))) {
      s = (1.0 / ((double)(n)));
      scale(a,n,s);
    };
  }

  public static void fft1drow(ComplexArray2D a,int n,int isign) {
    double pi = 3.14159265358979324;
    int log2n;
    int remainder;
    int point;
    int reverse;
    int row;
    int col;
    int i;
    int j;
    int k;
    int l;
    int stage;
    int lstar;
    Complex temp = new Complex(0.0,0.0);
    Complex tau = new Complex(0.0,0.0);
    Complex w = new Complex(0.0,0.0);
    Complex wp = new Complex(0.0,0.0);
    double theta;
    double s;
    temp.assign((new Complex(0.0,0.0)));
    tau.assign((new Complex(0.0,0.0)));
    w.assign((new Complex(0.0,0.0)));
    wp.assign((new Complex(0.0,0.0)));
    log2n = ((int)((0.5+(1.0/Math.log(2))*Math.log(((double)(n))))));
    for( point = 0; point <= (n - 1); point++) {
      remainder = point;
      reverse = 0;
      for( i = 1; i <= log2n; i++) {
        reverse = ((reverse * 2) + (remainder % 2));
        remainder = (remainder / 2);
      };
      if ((reverse > point)) {
        for( row = 0; row <= (n - 1); row++) {
          temp.assign(a.get(point,row));
          a.set(point,row,a.get(reverse,row));
          a.set(reverse,row,temp);
        };
      };
    };
    l = 1;
    for( stage = 1; stage <= log2n; stage++) {
      lstar = l;
      l = (l * 2);
      theta = (((2 * pi) * isign) / l);
      wp.assign((new Complex(((( -(2.0)) * (Math.sin((0.5 * theta)))) * (Math.sin((0.5 * theta)))),(Math.sin(theta)))));
      w.assign((new Complex(1.0,0.0)));
      for( k = 0; k <= (lstar - 1); k++) {
        for( i = k; i <= (n - 1); i += l) {
          j = (i + lstar);
          for( row = 0; row <= (n - 1); row++) {
            tau.assign((a.get(j,row).times(w)));
            a.set(j,row,(a.get(i,row).minus(tau)));
            a.set(i,row,(a.get(i,row).plus(tau)));
          };
        };
        w.assign((w.plus((w.times(wp)))));
      };
    };
    if ((isign == ( -(1)))) {
      s = (1.0 / ((double)(n)));
      scale(a,n,s);
    };
  }

  public static void scale(ComplexArray2D a,int n,double s) {
    int i;
    int j;
    Complex z = new Complex(0.0,0.0);
    z.assign((new Complex(0.0,0.0)));
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        a.set(j,i,(a.get(j,i).times(s)));
      };
    };
  }
}
