/* Demonstration program showing use of the Array package             */
/*                                                                    */
/* CHOLESKY: This program computes the Cholesky factor U of a         */
/* symmetric positive definite matrix A such that A = U'U where       */
/* U' is the transpose of U. U is an upper triangular matrix.         */

import com.ibm.math.array.*;	// import the Array package

public class cholesky {

  public static void main(String[] args) {
    int n = 1000;			      // problem size
    doubleArray2D A = new doubleArray2D(n,n); // n x n 2-dimensional array
    double t1;
    double t2;
    double time;
    double ops;
    double mflops;

    /*
     * Compute Cholesky factor using an optimized code.
     */
    initialize(A,n);
    t1 = (1e-3*System.currentTimeMillis());
    unroll(A,n);
    t2 = (1e-3*System.currentTimeMillis());
    time = (t2 - t1);
    ops = ((((1.0 * n) * n) * n) / 3.0);
    mflops = ((1.0e-6 * ops) / time);
    System.out.print("Optimized code, Time: "+time+" Mflops:"+mflops+" ");
    validate(A,n);

    /*
     * Compute Cholesky factor using a straightforward code.
     */
    initialize(A,n);
    t1 = (1e-3*System.currentTimeMillis());
    plain(A,n);
    t2 = (1e-3*System.currentTimeMillis());
    time = (t2 - t1);
    ops = ((((1.0 * n) * n) * n) / 3.0);
    mflops = ((1.0e-6 * ops) / time);
    System.out.print("Simple code, Time: "+time+" Mflops:"+mflops+" ");
    validate(A,n);
  }

  public static void initialize(doubleArray2D A,int n) {
    int i;
    int j;
    for( j = 0; j <= (n - 1); j++) {
      for( i = 0; i <= (n - 1); i++) {
        A.set(j,i,0);
      };
    };
    for( j = 0; j <= (n - 1); j++) {
      for( i = j; i <= (n - 1); i++) {
        A.set(i,j,(j + 1));
      };
    };
  }

  public static void validate(doubleArray2D A,int n) {
    int i;
    int j;
    int result;
    double eps;
    result = 1;
    eps = 1.0e-10;
    for( i = 0; i <= (n - 1); i++) {
      for( j = i; j <= (n - 1); j++) {
        if (((Math.abs((A.get(j,i) - 1.0))) > eps)) {
          result = 0;
        };
      };
    };
    if (result == 1) {
      System.out.println("Result is correct!");
    } else {
      System.out.println("Result is incorrect!");
    }
  }

  public static void plain(doubleArray2D A,int n) {
    int i;
    int j;
    int k;
    double s;
    double d;
    for( j = 0; j <= (n - 1); j++) {
      d = 0.0;
      for( k = 0; k <= (j - 1); k++) {
        s = 0.0;
        for( i = 0; i <= (k - 1); i++) {
          s = (s + (A.get(k,i) * A.get(j,i)));
        };
        s = ((A.get(j,k) - s) / A.get(k,k));
        A.set(j,k,s);
        d = (d + (s * s));
      };
      d = (A.get(j,j) - d);
      A.set(j,j,(Math.sqrt(d)));
    };
  }

  public static void unroll(doubleArray2D A,int n) {
    int i;
    int j;
    int k;
    double s;
    double d;
    double s0;
    double s1;
    double s2;
    double s3;
    double aij;
    for( j = 0; j <= (n - 1); j++) {
      d = 0.0;
      for( k = 0; k <= (j - 4); k += 4) {
        s0 = 0.0;
        s1 = 0.0;
        s2 = 0.0;
        s3 = 0.0;
        for( i = 0; i <= (k - 1); i++) {
          aij = A.get(j,i);
          s0 = (s0 + (A.get((k + 0),i) * aij));
          s1 = (s1 + (A.get((k + 1),i) * aij));
          s2 = (s2 + (A.get((k + 2),i) * aij));
          s3 = (s3 + (A.get((k + 3),i) * aij));
        };
        s0 = ((A.get(j,(k + 0)) - s0) / A.get((k + 0),(k + 0)));
        A.set(j,(k + 0),s0);
        s1 = (s1 + (A.get((k + 1),k) * A.get(j,k)));
        s1 = ((A.get(j,(k + 1)) - s1) / A.get((k + 1),(k + 1)));
        A.set(j,(k + 1),s1);
        s2 = ((s2 + (A.get((k + 2),k) * A.get(j,k))) + (A.get((k + 2),(k + 1)) * A.get(j,(k + 1))));
        s2 = ((A.get(j,(k + 2)) - s2) / A.get((k + 2),(k + 2)));
        A.set(j,(k + 2),s2);
        s3 = (((s3 + (A.get((k + 3),k) * A.get(j,k))) + (A.get((k + 3),(k + 1)) * A.get(j,(k + 1)))) + (A.get((k + 3),(k + 2)) * A.get(j,(k + 2))));
        s3 = ((A.get(j,(k + 3)) - s3) / A.get((k + 3),(k + 3)));
        A.set(j,(k + 3),s3);
        d = ((((d + (s0 * s0)) + (s1 * s1)) + (s2 * s2)) + (s3 * s3));
      };
      for( k = ((j / 4) * 4); k <= (j - 1); k++) {
        s = 0.0;
        for( i = 0; i <= (k - 1); i++) {
          s = (s + (A.get(k,i) * A.get(j,i)));
        };
        s = ((A.get(j,k) - s) / A.get(k,k));
        A.set(j,k,s);
        d = (d + (s * s));
      };
      d = (A.get(j,j) - d);
      A.set(j,j,(Math.sqrt(d)));
    };
  }
}
