/* Demonstration program showing use of the Array package             */
/*                                                                    */
/* SHALLOW: This is a shallow-water simulation code. It was adapted   */
/* from an NCAR (National Center for Atmospheric Research) code       */
/* originally in Fortran.                                             */

import com.ibm.math.array.*;	// import the Array package

public class shallow {

  public static void main(String[] args) {

    /*
     * Define problem parameters.
     */
    int n = 256;
    int size = (n + 2);
    int dt = 20;
    double dx = 0.25e+05;
    double dy = 0.25e+05;
    double alpha = 0.001;
    int n1 = (size - 2);
    int n2 = (size - 2);
    int nstep = dt;

    /*
     * Create problem arrays.
     */
    doubleArray2D u = new doubleArray2D(size,size);
    doubleArray2D v = new doubleArray2D(size,size);
    doubleArray2D p = new doubleArray2D(size,size);
    doubleArray2D uold = new doubleArray2D(size,size);
    doubleArray2D vold = new doubleArray2D(size,size);
    doubleArray2D pold = new doubleArray2D(size,size);
    doubleArray2D unew = new doubleArray2D(size,size);
    doubleArray2D vnew = new doubleArray2D(size,size);
    doubleArray2D pnew = new doubleArray2D(size,size);
    doubleArray2D cu = new doubleArray2D(size,size);
    doubleArray2D cv = new doubleArray2D(size,size);
    doubleArray2D z = new doubleArray2D(size,size);
    doubleArray2D h = new doubleArray2D(size,size);
    doubleArray2D psi = new doubleArray2D(size,size);

    double etime;
    double normu2;
    double flops;
    double floprt;
    double timer;
    int i;
    int j;
    int step;
    int result;
    double eps;
    double diff;
    eps = 1.0e-10;

    System.out.println("Grid size: "+n1+"x"+n2+" Time steps: "+nstep);
    
    /*
     * Initialize problem.
     */
    initial(size,u,v,p,uold,vold,pold,unew,vnew,pnew,cu,cv,z,h,psi,n1,n2,dx,dy,dt);

    /*
     * Do the computations.
     */
    etime = (1e-3*System.currentTimeMillis());
    for( step = 0; step <= (nstep - 1); step++) {
      calc1(size,u,v,p,uold,vold,pold,unew,vnew,pnew,cu,cv,z,h,psi,n1,n2,dx,dy);
      calc2(size,u,v,p,uold,vold,pold,unew,vnew,pnew,cu,cv,z,h,psi,n1,n2,dx,dy,dt);
      calc3(size,u,v,p,uold,vold,pold,unew,vnew,pnew,cu,cv,z,h,psi,step,n1,n2,alpha);
    };
    etime = ((1e-3*System.currentTimeMillis()) - etime);

    /*
     * Compute validation and performance parameters.
     */
    normu2 = 0;
    for( i = 1; i <= n1; i++) {
      for( j = 1; j <= n2; j++) {
        normu2 = (normu2 + (u.get(j,i) * v.get(j,i)));
      };
    };
    System.out.println("Validation parameter: "+normu2);

    flops = (((65 * n1) * n2) * nstep);
    floprt = 0;
    if ((etime > 0)) {
      floprt = ((1.0e-06 * flops) / etime);
    };
    System.out.println("Time: "+etime+" Mflops: "+floprt);

    diff = (Math.abs((normu2 + 2.37791268959808377)));
    if ((diff > eps)) {
      System.out.println("Result is incorrect!");
    } else {
      System.out.println("Result is correct!");
    };
  }

  public static void initial(int size,doubleArray2D u,doubleArray2D v,doubleArray2D p,doubleArray2D uold,doubleArray2D vold,doubleArray2D pold,doubleArray2D unew,doubleArray2D vnew,doubleArray2D pnew,doubleArray2D cu,doubleArray2D cv,doubleArray2D z,doubleArray2D h,doubleArray2D psi,int n1,int n2,double dx,double dy,int dt) {
    double tdt;
    double a;
    double el;
    double pi;
    double tpi;
    double di;
    double dj;
    double pcf;
    int i;
    int j;
    double idouble;
    double jdouble;
    tdt = dt;
    a = 1.0e6;
    el = (n1 * dx);
    pi = (4.0 * (Math.atan(1.)));
    tpi = (2.0 * pi);
    di = (tpi / n1);
    dj = (tpi / n2);
    pcf = ((((pi * pi) * a) * a) / (el * el));
    for( i = 1; i <= (n1 + 1); i++) {
      idouble = ((double)(i));
      for( j = 1; j <= (n2 + 1); j++) {
        jdouble = ((double)(j));
        psi.set(j,i,((a * (Math.sin(((idouble - 0.5) * di)))) * (Math.sin(((jdouble - 0.5) * dj)))));
        p.set(j,i,((pcf * ((Math.cos(((2. * (idouble - 1)) * di))) + (Math.cos(((2. * (jdouble - 1)) * dj))))) + 50000.0));
      };
    };
    for( i = 1; i <= n1; i++) {
      for( j = 1; j <= n2; j++) {
        u.set(j,i,(( -((psi.get((j + 1),(i + 1)) - psi.get(j,(i + 1))))) / dy));
        v.set(j,i,((psi.get((j + 1),(i + 1)) - psi.get((j + 1),i)) / dx));
      };
    };
    for( j = 0; j <= (n2 + 1); j++) {
      u.set(j,0,u.get(j,n1));
      v.set(j,(n1 + 1),v.get(j,1));
    };
    for( i = 0; i <= (n1 + 1); i++) {
      u.set((n2 + 1),i,u.get(1,i));
      v.set(0,i,v.get(n2,i));
    };
    for( i = 0; i <= (n1 + 1); i++) {
      for( j = 0; j <= (n2 + 1); j++) {
        uold.set(j,i,u.get(j,i));
        vold.set(j,i,v.get(j,i));
        pold.set(j,i,p.get(j,i));
      };
    };
  }

  public static void calc1(int size,doubleArray2D u,doubleArray2D v,doubleArray2D p,doubleArray2D uold,doubleArray2D vold,doubleArray2D pold,doubleArray2D unew,doubleArray2D vnew,doubleArray2D pnew,doubleArray2D cu,doubleArray2D cv,doubleArray2D z,doubleArray2D h,doubleArray2D psi,int n1,int n2,double dx,double dy) {
    double fsdx;
    double fsdy;
    int i;
    int j;
    fsdx = (4.0 / dx);
    fsdy = (4.0 / dy);
    for( j = 1; j <= n2; j++) {
      for( i = 1; i <= n1; i++) {
        cu.set(j,i,((0.5 * (p.get(j,(i + 1)) + p.get(j,i))) * u.get(j,i)));
        cv.set(j,i,((0.5 * (p.get((j + 1),i) + p.get(j,i))) * v.get(j,i)));
        z.set(j,i,(((fsdx * (v.get(j,(i + 1)) - v.get(j,i))) - (fsdy * (u.get((j + 1),i) - u.get(j,i)))) / (((p.get(j,i) + p.get(j,(i + 1))) + p.get((j + 1),(i + 1))) + p.get((j + 1),i))));
        h.set(j,i,(p.get(j,i) + (0.25 * ((((u.get(j,i) * u.get(j,i)) + (u.get(j,(i - 1)) * u.get(j,(i - 1)))) + (v.get(j,i) * v.get(j,i))) + (v.get((j - 1),i) * v.get((j - 1),i))))));
      };
    };
    for( j = 0; j <= (n2 + 1); j++) {
      cu.set(j,0,cu.get(j,n1));
      cv.set(j,(n1 + 1),cv.get(j,1));
      z.set(j,0,z.get(j,n1));
      h.set(j,(n1 + 1),h.get(j,1));
    };
    for( i = 0; i <= (n1 + 1); i++) {
      cu.set((n2 + 1),i,cu.get(1,i));
      cv.set(0,i,cv.get(n2,i));
      z.set(0,i,z.get(n2,i));
      h.set((n2 + 1),i,h.get(1,i));
    };
  }

  public static void calc2(int size,doubleArray2D u,doubleArray2D v,doubleArray2D p,doubleArray2D uold,doubleArray2D vold,doubleArray2D pold,doubleArray2D unew,doubleArray2D vnew,doubleArray2D pnew,doubleArray2D cu,doubleArray2D cv,doubleArray2D z,doubleArray2D h,doubleArray2D psi,int n1,int n2,double dx,double dy,int dt) {
    double tdts8;
    double tdtsdx;
    double tdtsdy;
    int i;
    int j;
    tdts8 = (dt / 8.0);
    tdtsdx = (dt / dx);
    tdtsdy = (dt / dy);
    for( j = 1; j <= n2; j++) {
      for( i = 1; i <= n1; i++) {
        unew.set(j,i,((uold.get(j,i) + ((tdts8 * (z.get(j,i) + z.get((j - 1),i))) * (((cv.get(j,(i + 1)) + cv.get(j,i)) + cv.get((j - 1),i)) + cv.get((j - 1),(i + 1))))) - (tdtsdx * (h.get(j,(i + 1)) - h.get(j,i)))));
        vnew.set(j,i,((vold.get(j,i) - ((tdts8 * (z.get(j,i) + z.get(j,(i - 1)))) * (((cu.get((j + 1),i) + cu.get((j + 1),(i - 1))) + cu.get(j,(i - 1))) + cu.get(j,i)))) - (tdtsdy * (h.get((j + 1),i) - h.get(j,i)))));
        pnew.set(j,i,((pold.get(j,i) - (tdtsdx * (cu.get(j,i) - cu.get(j,(i - 1))))) - (tdtsdy * (cv.get(j,i) - cv.get((j - 1),i)))));
        uold.set(j,i,unew.get(j,i));
        vold.set(j,i,vnew.get(j,i));
        pold.set(j,i,pnew.get(j,i));
      };
    };
  }

  public static void calc3(int size,doubleArray2D u,doubleArray2D v,doubleArray2D p,doubleArray2D uold,doubleArray2D vold,doubleArray2D pold,doubleArray2D unew,doubleArray2D vnew,doubleArray2D pnew,doubleArray2D cu,doubleArray2D cv,doubleArray2D z,doubleArray2D h,doubleArray2D psi,int step,int n1,int n2,double alpha) {
    int i;
    int j;
    if ((step <= 1)) {
      alpha = 0.0;
    };
    for( j = 1; j <= n2; j++) {
      for( i = 1; i <= n1; i++) {
        uold.set(j,i,(u.get(j,i) + (alpha * ((unew.get(j,i) - (2.0 * u.get(j,i))) + uold.get(j,i)))));
        vold.set(j,i,(v.get(j,i) + (alpha * ((vnew.get(j,i) - (2.0 * v.get(j,i))) + vold.get(j,i)))));
        pold.set(j,i,(p.get(j,i) + (alpha * ((pnew.get(j,i) - (2.0 * p.get(j,i))) + pold.get(j,i)))));
        u.set(j,i,unew.get(j,i));
        v.set(j,i,vnew.get(j,i));
        p.set(j,i,pnew.get(j,i));
      };
    };
    for( j = 0; j <= (n2 + 1); j++) {
      u.set(j,0,u.get(j,n1));
      v.set(j,(n1 + 1),v.get(j,1));
      p.set(j,(n1 + 1),p.get(j,1));
    };
    for( i = 0; i <= (n1 + 1); i++) {
      u.set((n2 + 1),i,u.get(1,i));
      v.set(0,i,v.get(n2,i));
      p.set((n2 + 1),i,p.get(1,i));
    };
  }
}
